==> TundereBird-0.28.tar.bz2: TundereBird-0.28/core/Smdn.Net.Imap4.Client/Smdn.Net.Imap4.Client.Session/ImapSession.CommandsNonAuthenticatedState.cs <==
// 
// Author:
//       smdn <smdn@mail.invisiblefulmoon.net>
// 
// Copyright (c) 2008-2010 smdn
// 
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// in the Software without restriction, including without limitation the rights
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
// 
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
// 
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
// THE SOFTWARE.

using System;
using System.Net;

using Smdn.Net.Imap4.Client.Transaction;
using Smdn.Net.Imap4.Client.Transaction.BuiltIn;
using Smdn.Net.MessageAccessProtocols;

namespace Smdn.Net.Imap4.Client.Session {
  partial class ImapSession {
    /*
     * transaction methods : non-authenticated state
     */

    /// <summary>sends STARTTLS command</summary>
    /// <remarks>valid in non-authenticated state</remarks>
    public ImapTransactionResult StartTls(UpgradeConnectionStreamCallback createAuthenticatedStreamCallback)
    {
      return StartTls(createAuthenticatedStreamCallback, false);
    }

    /// <summary>sends STARTTLS command</summary>
    /// <remarks>valid in non-authenticated state</remarks>
    public ImapTransactionResult StartTls(UpgradeConnectionStreamCallback createAuthenticatedStreamCallback, bool reissueCapability)
    {
      RejectNonConnectedState();

      if (state != ImapSessionState.NotAuthenticated)
        throw new ImapProtocolViolationException("already authenticated");

      if (createAuthenticatedStreamCallback == null)
        throw new ArgumentNullException("createAuthenticatedStreamCallback");

      ImapTransactionResult result;

      using (var t = new StartTlsTransaction(connection, createAuthenticatedStreamCallback)) {
        if ((result = ProcessTransaction(t, handlesIncapableAsException)).Succeeded)
          // 6.2.1. STARTTLS Command
          //       Once [TLS] has been started, the client MUST discard cached
          //       information about server capabilities and SHOULD re-issue the
          //       CAPABILITY command.
          SetServerCapabilities(new ImapCapabilityList());
        else
          return result;
      }

      if (reissueCapability)
        Capability();

      return result;
    }

    /// <summary>sends AUTHENTICATE command</summary>
    /// <remarks>valid in non-authenticated state</remarks>
    public ImapTransactionResult Authenticate(ICredentialsByHost credentials, ImapAuthenticationMechanism authenticationMechanism)
    {
      return Authenticate(credentials, null, authenticationMechanism);
    }

    /// <summary>sends AUTHENTICATE command</summary>
    /// <remarks>valid in non-authenticated state</remarks>
    public ImapTransactionResult Authenticate(ICredentialsByHost credentials, string username, ImapAuthenticationMechanism authenticationMechanism)
    {
      RejectNonConnectedState();

      if (state != ImapSessionState.NotAuthenticated)
        return new ImapTransactionResult(ImapTransactionResultCode.RequestDone,
                                         "already authenticated");

      if (credentials == null)
        throw new ArgumentNullException("credentials");
      if (authenticationMechanism == null)
        throw new ArgumentNullException("authenticationMechanism");

      var credential = credentials.LookupCredential(connection, username, authenticationMechanism);

      if (credential == null)
        return new ImapTransactionResult(ImapTransactionResultCode.RequestError,
                                         string.Format("credential not found for {0};AUTH={1}@{2}:{3}", username, authenticationMechanism, connection.Host, connection.Port));

      using (var t = new AuthenticateTransaction(connection, credential, serverCapabilities.Has(ImapCapability.SaslIR))) {
        t.Request.Arguments["authentication mechanism name"] = authenticationMechanism;

        if (ProcessTransaction(t, handlesIncapableAsException).Succeeded) {
          UpdateAuthority(credential.UserName, authenticationMechanism);
          TransitStateTo(ImapSessionState.Authenticated);
        }

        return t.Result;
      }
    }

    /// <summary>sends LOGIN command</summary>
    /// <remarks>valid in non-authenticated state</remarks>
    public ImapTransactionResult Login(ICredentialsByHost credentials)
    {
      return Login(credentials, null);
    }

    /// <summary>sends LOGIN command</summary>
    /// <remarks>valid in non-authenticated state</remarks>
    public ImapTransactionResult Login(ICredentialsByHost credentials, string username)
    {
      RejectNonConnectedState();

      if (state != ImapSessionState.NotAuthenticated)
        return new ImapTransactionResult(ImapTransactionResultCode.RequestDone,
                                         "already logged in");

      // RFC 2595 Using TLS with IMAP, POP3 and ACAP
      // http://tools.ietf.org/html/rfc2595
      // 3.2. IMAP LOGINDISABLED capability
      //    An IMAP client which complies with this specification MUST NOT issue
      //    the LOGIN command if this capability is present.
      if (serverCapabilities.Has(Imap4.ImapCapability.LoginDisabled))
        throw new ImapIncapableException("LOGIN is disabled");

      if (credentials == null)
        throw new ArgumentNullException("credentials");

      var credential = credentials.LookupCredential(connection, username, null);

      if (credential == null)
        return new ImapTransactionResult(ImapTransactionResultCode.RequestError,
                                         string.Format("credential not found for {0}@{1}:{2}", username, connection.Host, connection.Port));

      using (var t = new LoginTransaction(connection, credential)) {
        if (ProcessTransaction(t, handlesIncapableAsException).Succeeded) {
          UpdateAuthority(credential.UserName, null);
          TransitStateTo(ImapSessionState.Authenticated);
        }

        return t.Result;
      }
    }

    internal void UpdateAuthority(string username, ImapAuthenticationMechanism authType)
    {
      // TODO: use ImapUriBuilder
      var scheme = string.Format("{0}://", connection.IsSecurePortConnection ? ImapUri.UriSchemeImaps : ImapUri.UriSchemeImap);
      var userInfo = (username == null)
        ? string.Empty
        : (authType == null)
          ? string.Format("{0}@", username)
          : string.Format("{0};AUTH={1}@", username, authType);

      authority = new Uri(string.Format("{0}{1}{2}:{3}", scheme, userInfo, connection.Host, connection.Port));

      TraceInfo("authority: {0}", authority);
    }
  }
}

==> TundereBird-0.29.tar.bz2: TundereBird-0.29/core/Smdn.Net.Imap4.Client/Smdn.Net.Imap4.Client.Session/ImapSession.CommandsNonAuthenticatedState.cs <==
// 
// Author:
//       smdn <smdn@mail.invisiblefulmoon.net>
// 
// Copyright (c) 2008-2010 smdn
// 
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// in the Software without restriction, including without limitation the rights
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
// 
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
// 
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
// THE SOFTWARE.

using System;
using System.Net;

using Smdn.Collections;
using Smdn.Net.Imap4.Client.Transaction;
using Smdn.Net.Imap4.Client.Transaction.BuiltIn;
using Smdn.Net.Imap4.Protocol.Client;

namespace Smdn.Net.Imap4.Client.Session {
  partial class ImapSession {
    /*
     * transaction methods : non-authenticated state
     */

    /// <summary>sends STARTTLS command</summary>
    /// <remarks>valid in non-authenticated state</remarks>
    public ImapTransactionResult StartTls(UpgradeConnectionStreamCallback createAuthenticatedStreamCallback)
    {
      return StartTls(createAuthenticatedStreamCallback, false);
    }

    /// <summary>sends STARTTLS command</summary>
    /// <remarks>valid in non-authenticated state</remarks>
    public ImapTransactionResult StartTls(UpgradeConnectionStreamCallback createAuthenticatedStreamCallback, bool reissueCapability)
    {
      RejectNonConnectedState();

      if (state != ImapSessionState.NotAuthenticated)
        throw new ImapProtocolViolationException("already authenticated");

      if (createAuthenticatedStreamCallback == null)
        throw new ArgumentNullException("createAuthenticatedStreamCallback");

      ImapTransactionResult result;

      using (var t = new StartTlsTransaction(connection, createAuthenticatedStreamCallback)) {
        if ((result = ProcessTransaction(t, handlesIncapableAsException)).Succeeded)
          // 6.2.1. STARTTLS Command
          //       Once [TLS] has been started, the client MUST discard cached
          //       information about server capabilities and SHOULD re-issue the
          //       CAPABILITY command.
          SetServerCapabilities(new ImapCapabilityList());
        else
          return result;
      }

      if (reissueCapability)
        Capability();

      return result;
    }

    /// <summary>sends AUTHENTICATE command</summary>
    /// <remarks>valid in non-authenticated state</remarks>
    public ImapTransactionResult Authenticate(ICredentialsByHost credentials,
                                              ImapAuthenticationMechanism authenticationMechanism)
    {
      return Authenticate(credentials, authenticationMechanism, false);
    }

    /// <summary>sends AUTHENTICATE command</summary>
    /// <remarks>valid in non-authenticated state</remarks>
    public ImapTransactionResult Authenticate(ICredentialsByHost credentials,
                                              ImapAuthenticationMechanism authenticationMechanism,
                                              bool reissueCapability)
    {
      return Authenticate(credentials, null, authenticationMechanism, reissueCapability);
    }

    /// <summary>sends AUTHENTICATE command</summary>
    /// <remarks>valid in non-authenticated state</remarks>
    public ImapTransactionResult Authenticate(ICredentialsByHost credentials,
                                              string username,
                                              ImapAuthenticationMechanism authenticationMechanism)
    {
      return Authenticate(credentials, username, authenticationMechanism, false);
    }

    /// <summary>sends AUTHENTICATE command</summary>
    /// <remarks>valid in non-authenticated state</remarks>
    public ImapTransactionResult Authenticate(ICredentialsByHost credentials,
                                              string username,
                                              ImapAuthenticationMechanism authenticationMechanism,
                                              bool reissueCapability)
    {
      RejectNonConnectedState();

      if (state != ImapSessionState.NotAuthenticated)
        return new ImapTransactionResult(ImapTransactionResultCode.RequestDone,
                                         "already authenticated");

      if (credentials == null)
        throw new ArgumentNullException("credentials");
      if (authenticationMechanism == null)
        throw new ArgumentNullException("authenticationMechanism");

      var credential = credentials.LookupCredential(connection, username, authenticationMechanism);

      if (credential == null)
        return new ImapTransactionResult(ImapTransactionResultCode.RequestError,
                                         string.Format("credential not found for {0};AUTH={1}@{2}:{3}", username, authenticationMechanism, connection.Host, connection.Port));

      using (var t = new AuthenticateTransaction(connection, credential, serverCapabilities.Has(ImapCapability.SaslIR))) {
        t.Request.Arguments["authentication mechanism name"] = authenticationMechanism;

        return AuthenticateInternal(t, credential.UserName, authenticationMechanism, reissueCapability);
      }
    }

    /// <summary>sends LOGIN command</summary>
    /// <remarks>valid in non-authenticated state</remarks>
    public ImapTransactionResult Login(ICredentialsByHost credentials)
    {
      return Login(credentials, false);
    }

    /// <summary>sends LOGIN command</summary>
    /// <remarks>valid in non-authenticated state</remarks>
    public ImapTransactionResult Login(ICredentialsByHost credentials,
                                       bool reissueCapability)
    {
      return Login(credentials, null, reissueCapability);
    }

    /// <summary>sends LOGIN command</summary>
    /// <remarks>valid in non-authenticated state</remarks>
    public ImapTransactionResult Login(ICredentialsByHost credentials,
                                       string username)
    {
      return Login(credentials, username, false);
    }

    /// <summary>sends LOGIN command</summary>
    /// <remarks>valid in non-authenticated state</remarks>
    public ImapTransactionResult Login(ICredentialsByHost credentials,
                                       string username,
                                       bool reissueCapability)
    {
      RejectNonConnectedState();

      if (state != ImapSessionState.NotAuthenticated)
        return new ImapTransactionResult(ImapTransactionResultCode.RequestDone,
                                         "already logged in");

      // RFC 2595 Using TLS with IMAP, POP3 and ACAP
      // http://tools.ietf.org/html/rfc2595
      // 3.2. IMAP LOGINDISABLED capability
      //    An IMAP client which complies with this specification MUST NOT issue
      //    the LOGIN command if this capability is present.
      if (serverCapabilities.Has(Imap4.ImapCapability.LoginDisabled))
        throw new ImapIncapableException("LOGIN is disabled");

      if (credentials == null)
        throw new ArgumentNullException("credentials");

      var credential = credentials.LookupCredential(connection, username, null);

      if (credential == null)
        return new ImapTransactionResult(ImapTransactionResultCode.RequestError,
                                         string.Format("credential not found for {0}@{1}:{2}", username, connection.Host, connection.Port));

      using (var t = new LoginTransaction(connection, credential)) {
        return AuthenticateInternal(t, credential.UserName, null, reissueCapability);
      }
    }

    private ImapTransactionResult AuthenticateInternal(IImapTransaction t,
                                                       string username,
                                                       ImapAuthenticationMechanism authenticationMechanism,
                                                       bool reissueCapability)
    {
      var result = ProcessTransaction(t, handlesIncapableAsException);

      if (result.Succeeded) {
        UpdateAuthority(username, authenticationMechanism);
        TransitStateTo(ImapSessionState.Authenticated);

        if (result.ReceivedResponses.Exists(ImapCapabilityResponseProcessor.IsCapabilityResponse))
          reissueCapability = false;
      }
      else {
        return result;
      }

      if (reissueCapability)
        Capability();

      return result;
    }

    internal void UpdateAuthority(string username, ImapAuthenticationMechanism authType)
    {
      authority.Scheme    = connection.IsSecurePortConnection ? ImapUri.UriSchemeImaps : ImapUri.UriSchemeImap;
      authority.Host      = connection.Host;
      authority.Port      = connection.Port;
      authority.UserName  = username;
      authority.AuthType  = authType;

      TraceInfo("authority: {0}", authority);
    }
  }
}
